@@ -11,12 +11,23 @@ import re
 import os
 import os.path
 import sys
+import glob
 import textwrap
 import datetime
 import collections
-
 from subprocess import Popen, PIPE
 
+try:
+    import pystache
+except ImportError:
+    pystache = None
+
+try:
+    import mako
+except ImportError:
+    mako = None
+
+
 
 usage_msg = """usage: %(exname)s [REPOS]
 
@@ -50,10 +61,18 @@ def die(msg=None):
 ##
 
 
+_config_env = {}
+
+
+def available_in_config(f):
+    _config_env[f.__name__] = f
+    return f
+
+
 def load_config_file(filename, fail_if_not_present=True):
     """Loads data from a config file."""
 
-    config = {}
+    config = _config_env.copy()
 
     if os.path.exists(filename):
         try:
@@ -393,34 +412,6 @@ class GitRepos(object):
         raise NotImplementedError("Unsupported getitem %r object." % key)
 
 
-##
-## The actual changelog code
-##
-
-
-def make_section_string(title, sections, section_label_order):
-    s = ""
-    if len(sections) != 0:
-        title = title.strip()
-        s += title + "\n"
-        s += "-" * len(title) + "\n\n"
-
-        nb_sections = len(sections)
-        for section in section_label_order:
-            if section not in sections:
-                continue
-
-            section_label = section if section else "Other"
-
-            if not (section_label == "Other" and nb_sections == 1):
-                s += section_label + "\n"
-                s += "~" * len(section_label) + "\n\n"
-
-            for entry in sections[section]:
-                s += entry
-    return s
-
-
 def first_matching(section_regexps, string):
     for section, regexps in section_regexps:
         if regexps is None:
@@ -429,6 +420,158 @@ def first_matching(section_regexps, string):
             if re.search(regexp, string) is not None:
                 return section
 
+##
+## Output Engines
+##
+
+@available_in_config
+def rest_py(data, opts={}):
+    """Returns ReStructured Text changelog content from data"""
+
+    def rest_title(label, char="="):
+        return (label.strip() + "\n") + (char * len(label) + "\n")
+
+    def render_version(title, sections):
+        s = rest_title(title, char="-") + "\n"
+
+        nb_sections = len(sections)
+        for section in sections:
+
+            section_label = section["label"] if section.get("label", None) \
+                            else "Other"
+
+            if not (section_label == "Other" and nb_sections == 1):
+                s += rest_title(section_label, "~") + "\n"
+
+            for commit in section["commits"]:
+                s += render_commit(commit)
+        return s
+
+    def render_commit(commit, opts=opts):
+        subject = commit["subject"]
+        subject += " [%s]" % (commit["author"], )
+
+        entry = indent('\n'.join(textwrap.wrap(ucfirst(subject))),
+                       first="- ").strip() + "\n\n"
+
+        if commit["body"]:
+            entry += indent(paragraph_wrap(commit["body"],
+                                           regexp=opts["body_split_regexp"]))
+            entry += "\n\n"
+
+        return entry
+
+
+    return (rest_title(data["title"], char="=") + "\n" +
+            "".join(render_version(title=version["label"],
+                                   sections=version["sections"])
+                    for version in data["versions"]
+                    if len(version["sections"]) > 0))
+
+## formatter engines
+
+if pystache:
+
+    @available_in_config
+    def mustache(template_name):
+        """Return a callable that will render a changelog data structure
+
+        returned callable must take 2 arguments ``data`` and ``opts``.
+
+        """
+        template_dir = os.path.join(
+            os.path.dirname(os.path.realpath(__file__)),
+            "templates", "mustache")
+
+        template_path = os.path.join(template_dir, "%s.tpl" % template_name)
+
+        if not os.path.exists(template_path):
+            templates = glob.glob(os.path.join(template_dir, "*.tpl"))
+            if len(templates) > 0:
+                msg = "Available mustache templates:\n - " + \
+                      " - ".join(os.path.basename(f).split(".")[-1]
+                                 for f in templates)
+            else:
+                msg = "No available mustache templates found in %r." \
+                      % template_dir
+            die("%s\n" % msg +
+                "Invalid mustache template name %s." % template_name)
+
+        with open(template_path) as f:
+            template = f.read()
+
+        def renderer(data, opts):
+
+            ## mustache is very simple so we need to add some intermediate
+            ## values
+            data["title_chars"] = list(data["title"])
+            for version in data["versions"]:
+                version["label_chars"] = list(version["label"])
+                for section in version["sections"]:
+                    section["label_chars"] = list(section["label"])
+                    for commit in section["commits"]:
+                        commit["body_indented"] = indent(
+                            paragraph_wrap(commit["body"],
+                                           regexp=opts["body_split_regexp"]),
+                            chars="    ")
+
+            return pystache.render(template, data)
+
+        return renderer
+
+else:
+
+    @available_in_config
+    def mustache(template_name):
+        die("Required 'pystache' python module not found.")
+
+
+if mako:
+
+    import mako.template
+
+    mako_env = dict((f.__name__, f) for f in (ucfirst, indent, textwrap,
+                                              paragraph_wrap))
+
+    @available_in_config
+    def makotemplate(template_name):
+        """Return a callable that will render a changelog data structure
+
+        returned callable must take 2 arguments ``data`` and ``opts``.
+
+        """
+        template_dir = os.path.join(
+            os.path.dirname(os.path.realpath(__file__)),
+            "templates", "mako")
+
+        template_path = os.path.join(template_dir, "%s.tpl" % template_name)
+
+        if not os.path.exists(template_path):
+            print "Available mako templates:\n - " + \
+                  " - ".join(os.path.basename(f).split(".")[-1]
+                             for f in glob.glob(os.path.join(template_dir,
+                                                             "*.tpl")))
+            die("No %r a valid mako template name." % template_name)
+
+        template = mako.template.Template(filename=template_path)
+        def renderer(data, opts):
+            kwargs = mako_env.copy()
+            kwargs.update({"data": data,
+                           "opts": opts})
+            return template.render(**kwargs)
+
+        return renderer
+
+else:
+
+    @available_in_config
+    def makotemplate(template_name):
+        die("Required 'mako' python module not found.")
+
+
+##
+## Data Structure
+##
 
 def changelog(repository,
               ignore_regexps=[],
@@ -437,10 +580,52 @@ def changelog(repository,
               unreleased_version_label="unreleased",
               tag_filter_regexp=r"\d+\.\d+(\.\d+)?",
               body_split_regexp="\n\n",
+              output_engine=rest_py,
               ):
+    """Returns a string containing the changelog of given repository
 
-    s = "Changelog\n"
-    s += "=========\n\n"
+    This function returns a string corresponding to the template rendered with
+    the changelog data tree.
+
+    (see ``gitchangelog.rc.sample`` file for more info)
+
+    :param repository: target ``GitRepos`` object
+    :param ignore_regexps: list of regexp identifying ignored commit messages
+    :param replace_regexps: regexps used to replace elements in commit messages
+    :param section_regexps: regexps identifying sections
+    :param tag_filter_regexp: regexp to match tags used as version
+    :param body_split_regexp: regexp identifying the body of a commit message
+    :param unreleased_version_label: version label for untagged commits
+    :param template_format: format of template to generate the changelog
+
+    :returns: content of changelog
+
+    """
+
+    def new_version(tag, date, opts) :
+        title = "%s (%s)" % (tag, date) if tag else \
+                opts["unreleased_version_label"]
+        return {"label": title,
+                "tag": tag,
+                "date": date,
+                }
+
+    opts = {
+        'unreleased_version_label': unreleased_version_label,
+        'body_split_regexp': body_split_regexp,
+        }
+
+    # setting main container of changelog elements
+    title = "Changelog"
+    changelog = {"title": title,
+                 "versions": []}
+
+    ## Hash to speedup lookups
+    versions_done = {}
+
+    ## Create unreleased version
+    current_version = new_version(tag=None, date=None, opts=opts)
+    sections = collections.defaultdict(list)
 
     tags = [tag
             for tag in reversed(repository.tags)
@@ -448,28 +633,27 @@ def changelog(repository,
 
     section_order = [k for k, _v in section_regexps]
 
-    title = unreleased_version_label + "\n"
-    sections = collections.defaultdict(list)
-
     for commit in reversed(repository[:]):
 
         tags_of_commit = [tag for tag in tags
-                         if tag == commit]
+                          if tag == commit]
+
         if len(tags_of_commit) > 0:
             tag = tags_of_commit[0]
-            ## End of sections, let's flush current one.
-            s += make_section_string(title, sections, section_order)
 
-            title = "%s (%s)\n" % (tag.identifier, commit.date)
+            ## End of current version, let's flush current one.
+            current_version["sections"] = [{"label": k, "commits": sections[k]}
+                                           for k in section_order
+                                           if k in sections]
+            changelog["versions"].append(current_version)
+            versions_done[tag] = current_version
+            current_version = new_version(tag.identifier, commit.date, opts)
             sections = collections.defaultdict(list)
 
-        ## Ignore some commit subject
-        if any([re.search(pattern, commit.subject) is not None
-                for pattern in ignore_regexps]):
+        if any(re.search(pattern, commit.subject) is not None
+               for pattern in ignore_regexps):
             continue
 
-        ## Put message in sections if possible
-
         matched_section = first_matching(section_regexps, commit.subject)
 
         ## Replace content in commit subject
@@ -478,22 +662,24 @@ def changelog(repository,
         for regexp, replacement in replace_regexps.iteritems():
             subject = re.sub(regexp, replacement, subject)
 
-        ## Finaly print out the commit
+        ## Finally storing the commit in the matching section
 
         subject = final_dot(subject)
-        subject += " [%s]" % (commit.author_name, )
-        entry = indent('\n'.join(textwrap.wrap(ucfirst(subject))),
-                       first="- ").strip() + "\n\n"
 
-        if commit.body:
-            entry += indent(paragraph_wrap(commit.body,
-                                           regexp=body_split_regexp))
-            entry += "\n\n"
+        sections[matched_section].append({
+            "author": commit.author_name,
+            "subject": subject,
+            "body": commit.body
+            })
 
-        sections[matched_section].append(entry)
+    ## Don't forget last commits:
+    current_version["sections"] = [{"label": k, "commits": sections[k]}
+                                   for k in section_order
+                                   if k in sections]
+    changelog["versions"].append(current_version)
+
+    return output_engine(data=changelog, opts=opts)
 
-    s += make_section_string(title, sections, section_order)
-    return s
 
 ##
 ## Main
@@ -535,16 +721,14 @@ def main():
         if changelogrc:
             if not os.path.exists(changelogrc):
                 if enforce_file_existence:
-                    print "File %r does not exists." % changelogrc
-                    sys.exit(1)
+                    die("File %r does not exists." % changelogrc)
                 else:
                     continue  ## changelogrc valued, but file does not exists
             else:
                 break
 
     if not changelogrc or not os.path.exists(changelogrc):
-        print "Not %s config file found anywhere." % basename
-        sys.exit(1)
+        die("No %s config file found anywhere." % basename)
 
     config = load_config_file(os.path.expanduser(changelogrc))
 
@@ -555,6 +739,7 @@ def main():
         unreleased_version_label=config['unreleased_version_label'],
         tag_filter_regexp=config['tag_filter_regexp'],
         body_split_regexp=config['body_split_regexp'],
+        output_engine=config.get("output_engine", rest_py),
     )
 
 ##
